import {
	ClassType,
	PropertyType,
} from '@loafer/core/constants/types';

import {
	Assert,
} from '@loafer/core/util';

import {
  PouchMetadataElement,
} from '@loafer/pouches';

import {
  HierarchicalPouchFactory,
  PouchFactory,
} from '@loafer/pouches/factory';

import {
  PouchDefinition,
  PouchExpressionResolver,
  SingletonPouchRegistry,
} from '@loafer/pouches/factory/config';




export class ConstructorArgumentValues {
	private readonly indexedArgumentValues: Map<number, ValueHolder> = new Map();
	private readonly genericArgumentValues: ValueHolder[] = [];

	/**
	 * Deep copy constructor.
	 * @param original the ConstructorArgumentValues to copy
	 */
	public constructor(original?: ConstructorArgumentValues) {
		this.addArgumentValues(original);
	}


	/**
	 * Copy all given argument values into this object, using separate holder
	 * instances to keep the values independent from the original object.
	 * <p>Note: Identical ValueHolder instances will only be registered once,
	 * to allow for merging and re-merging of argument value definitions. Distinct
	 * ValueHolder instances carrying the same content are of course allowed.
	 */
	public addArgumentValues(other: ConstructorArgumentValues): void {
		if (other != null) {
      other.indexedArgumentValues.forEach((valueHolder, key, map) => {
        this.addOrMergeIndexedArgumentValue(key, valueHolder.copy());
      });
      other.genericArgumentValues.filter((valueHolder, index, array) => {
        return -1 === this.genericArgumentValues.indexOf(valueHolder);
      }).forEach((valueHolder, index, array) => {
        this.addOrMergeGenericArgumentValue(valueHolder.copy());
      });
		}
	}


	/**
	 * Add an argument value for the given index in the constructor argument list.
	 * @param index the index in the constructor argument list
	 * @param newValue the argument value in the form of a ValueHolder
	 */
  public addIndexedArgumentValue(index: number, value: ValueHolder): void;
	/**
	 * Add an argument value for the given index in the constructor argument list.
	 * @param index the index in the constructor argument list
	 * @param value the argument value
	 * @param type the type of the constructor argument
	 */
	public addIndexedArgumentValue(index: number, value: any, type?: ClassType): void {
    Assert.isTrue(index >= 0, 'Index must not be negative');

    let newValue: ValueHolder;
    if (!(value instanceof ValueHolder)) {
      newValue = new ValueHolder(value, type);
    }
    Assert.notNull(newValue, 'ValueHolder must not be null');
    this.addOrMergeIndexedArgumentValue(index, newValue);
	}

	/**
	 * Add an argument value for the given index in the constructor argument list,
	 * merging the new value (typically a collection) with the current value
	 * if demanded: see {@link org.springframework.beans.Mergeable}.
	 * @param key the index in the constructor argument list
	 * @param newValue the argument value in the form of a ValueHolder
	 */
	private addOrMergeIndexedArgumentValue(key: number, newValue: ValueHolder): void {
		let currentValue: ValueHolder = this.indexedArgumentValues.get(key);
		if (currentValue !== undefined && newValue.getValue() instanceof Mergeable) {
			let mergeable: Mergeable = <Mergeable>newValue.getValue();
			if (mergeable.isMergeEnabled()) {
				newValue.setValue(mergeable.merge(currentValue.getValue()));
			}
		}
		this.indexedArgumentValues.set(key, newValue);
	}

	/**
	 * Check whether an argument value has been registered for the given index.
	 * @param index the index in the constructor argument list
	 */
  public hasIndexedArgumentValue(index: number): boolean {
    return this.indexedArgumentValues.has(index);
  }

	/**
	 * Get argument value for the given index in the constructor argument list.
	 * @param index the index in the constructor argument list
	 * @param requiredType the type to match (can be {@code null} to match
	 * untyped values only)
	 * @param requiredName the type to match (can be {@code null} to match
	 * unnamed values only, or empty String to match any name)
	 * @return the ValueHolder for the argument, or {@code null} if none set
	 */
	public getIndexedArgumentValue(index: number,  requiredType: ClassType, requiredName?: PropertyType): ValueHolder {
		Assert.isTrue(index >= 0, 'Index must not be negative');
		let valueHolder: ValueHolder = this.indexedArgumentValues.get(index);
		if (undefined !== valueHolder &&
				(valueHolder.getType() == null ||
						(requiredType != null && ClassUtils.matchesTypeName(requiredType, valueHolder.getType()))) &&
				(valueHolder.getName() == null || "".equals(requiredName) ||
						(requiredName != null && requiredName.equals(valueHolder.getName())))) {
			return valueHolder;
		}
		return null;
	}

	/**
	 * Return the map of indexed argument values.
	 * @return unmodifiable Map with Integer index as key and ValueHolder as value
	 * @see ValueHolder
	 */
	public getIndexedArgumentValues(): Map<number, ValueHolder> {
		return Collections.unmodifiableMap(this.indexedArgumentValues);
	}


	/**
	 * Add a generic argument value to be matched by type.
	 * <p>Note: A single generic argument value will just be used once,
	 * rather than matched multiple times.
	 * @param value the argument value
	 */
	public addGenericArgumentValue(value: any): void {
		this.genericArgumentValues.add(new ValueHolder(value));
	}

	/**
	 * Add a generic argument value to be matched by type.
	 * <p>Note: A single generic argument value will just be used once,
	 * rather than matched multiple times.
	 * @param value the argument value
	 * @param type the type of the constructor argument
	 */
	public addGenericArgumentValue(value: any, type: string): void {
		this.genericArgumentValues.add(new ValueHolder(value, type));
	}

	/**
	 * Add a generic argument value to be matched by type or name (if available).
	 * <p>Note: A single generic argument value will just be used once,
	 * rather than matched multiple times.
	 * @param newValue the argument value in the form of a ValueHolder
	 * <p>Note: Identical ValueHolder instances will only be registered once,
	 * to allow for merging and re-merging of argument value definitions. Distinct
	 * ValueHolder instances carrying the same content are of course allowed.
	 */
	public addGenericArgumentValue(newValue: ValueHolder): void {
		Assert.notNull(newValue, "ValueHolder must not be null");
		if (!this.genericArgumentValues.contains(newValue)) {
			addOrMergeGenericArgumentValue(newValue);
		}
	}

	/**
	 * Add a generic argument value, merging the new value (typically a collection)
	 * with the current value if demanded: see {@link org.springframework.beans.Mergeable}.
	 * @param newValue the argument value in the form of a ValueHolder
	 */
	private addOrMergeGenericArgumentValue(newValue: ValueHolder): void {
		if (newValue.getName() != null) {
			for (Iterator <ValueHolder> it = this.genericArgumentValues.iterator(); it.hasNext();) {
				ValueHolder currentValue = it.next();
				if (newValue.getName().equals(currentValue.getName())) {
					if (newValue.getValue() instanceof Mergeable) {
						Mergeable; mergeable = (Mergeable) newValue.getValue();
						if (mergeable.isMergeEnabled()) {
							newValue.setValue(mergeable.merge(currentValue.getValue()));
						}
					}
					it.remove();
				}
			}
		}
		this.genericArgumentValues.add(newValue);
	}

	/**
	 * Look for a generic argument value that matches the given type.
	 * @param requiredType the type to match
	 * @return the ValueHolder for the argument, or {@code null} if none set
	 */

	public getGenericArgumentValue(requiredType: ClassType): ValueHolder {
		return getGenericArgumentValue(requiredType, null, null);
	}

	/**
	 * Look for a generic argument value that matches the given type.
	 * @param requiredType the type to match
	 * @param requiredName the name to match
	 * @return the ValueHolder for the argument, or {@code null} if none set
	 */
	
	public getGenericArgumentValue(requiredType: ClassType, requiredName: PropertyType): ValueHolder {
		return getGenericArgumentValue(requiredType, requiredName, null);
	}

	/**
	 * Look for the next generic argument value that matches the given type,
	 * ignoring argument values that have already been used in the current
	 * resolution process.
	 * @param requiredType the type to match (can be {@code null} to find
	 * an arbitrary next generic argument value)
	 * @param requiredName the name to match (can be {@code null} to not
	 * match argument values by name, or empty String to match any name)
	 * @param usedValueHolders a Set of ValueHolder objects that have already been used
	 * in the current resolution process and should therefore not be returned again
	 * @return the ValueHolder for the argument, or {@code null} if none found
	 */

	public getGenericArgumentValue(requiredType: ClassType,  requiredName: string,  usedValueHolders: Set<ValueHolder>): ValueHolder {
		for (ValueHolder valueHolder : this.genericArgumentValues) {
			if (usedValueHolders != null && usedValueHolders.contains(valueHolder)) {
				continue;
			}
			if (valueHolder.getName() != null && !''.equals(requiredName) &&
					(requiredName == null || !valueHolder.getName().equals(requiredName))) {
				continue;
			}
			if (valueHolder.getType() != null &&
					(requiredType == null || !ClassUtils.matchesTypeName(requiredType, valueHolder.getType()))) {
				continue;
			}
			if (requiredType != null && valueHolder.getType() == null && valueHolder.getName() == null &&
					!ClassUtils.isAssignableValue(requiredType, valueHolder.getValue())) {
				continue;
			}
			return valueHolder;
		}
		return null;
	}

	/**
	 * Return the list of generic argument values.
	 * @return unmodifiable List of ValueHolders
	 * @see ValueHolder
	 */
	public getGenericArgumentValues(): ValueHolder[] {
		return Collections.unmodifiableList(this.genericArgumentValues);
	}


	/**
	 * Look for an argument value that either corresponds to the given index
	 * in the constructor argument list or generically matches by type.
	 * @param index the index in the constructor argument list
	 * @param requiredType the parameter type to match
	 * @return the ValueHolder for the argument, or {@code null} if none set
	 */

	public getArgumentValue(index: number, requiredType: ClassType): ValueHolder {
		return getArgumentValue(index, requiredType, null, null);
	}

	/**
	 * Look for an argument value that either corresponds to the given index
	 * in the constructor argument list or generically matches by type.
	 * @param index the index in the constructor argument list
	 * @param requiredType the parameter type to match
	 * @param requiredName the parameter name to match
	 * @return the ValueHolder for the argument, or {@code null} if none set
	 */

	public getArgumentValue(index: number, requiredType: ClassType, requiredName: string): ValueHolder {
		return getArgumentValue(index, requiredType, requiredName, null);
	}

	/**
	 * Look for an argument value that either corresponds to the given index
	 * in the constructor argument list or generically matches by type.
	 * @param index the index in the constructor argument list
	 * @param requiredType the parameter type to match (can be {@code null}
	 * to find an untyped argument value)
	 * @param requiredName the parameter name to match (can be {@code null}
	 * to find an unnamed argument value, or empty String to match any name)
	 * @param usedValueHolders a Set of ValueHolder objects that have already
	 * been used in the current resolution process and should therefore not
	 * be returned again (allowing to return the next generic argument match
	 * in case of multiple generic argument values of the same type)
	 * @return the ValueHolder for the argument, or {@code null} if none set
	 */

	public getArgumentValue(index: number,  requiredType: ClassType,  requiredName: string,  usedValueHolders: Set<ValueHolder>): ValueHolder {
		Assert.isTrue(index >= 0, 'Index must not be negative');
		ValueHolder; valueHolder = getIndexedArgumentValue(index, requiredType, requiredName);
		if (valueHolder == null) {
			valueHolder = getGenericArgumentValue(requiredType, requiredName, usedValueHolders);
		}
		return valueHolder;
	}

	/**
	 * Return the number of argument values held in this instance,
	 * counting both indexed and generic argument values.
	 */
	public getArgumentCount(): number {
		return (this.indexedArgumentValues.size() + this.genericArgumentValues.size());
	}

	/**
	 * Return if this holder does not contain any argument values,
	 * neither indexed ones nor generic ones.
	 */
	public isEmpty(): boolean {
		return (this.indexedArgumentValues.isEmpty() && this.genericArgumentValues.isEmpty());
	}

	/**
	 * Clear this holder, removing all argument values.
	 */
	public clear(): void {
		this.indexedArgumentValues.clear();
		this.genericArgumentValues.clear();
	}

}


/**
 * Holder for a constructor argument value, with an optional type
 * attribute indicating the target type of the actual constructor argument.
 */
export class ValueHolder implements PouchMetadataElement {
  private value: any;


  private type: string;


  private name: string;


  private source: any;

  private converted: boolean = false;


  private convertedValue: any;

  /**
   * Create a new ValueHolder for the given value, type and name.
   * @param value the argument value
   * @param type the type of the constructor argument
   * @param name the name of the constructor argument
   */
  public constructor( value: any,  type?: string,  name?: string) {
    this.value = value;
    this.type = type;
    this.name = name;
  }

  /**
   * Set the value for the constructor argument.
   * @see PropertyPlaceholderConfigurer
   */
  public set Value(value: any) {
    this.value = value;
  }

  /**
   * Return the value for the constructor argument.
   */

  public get Value(): any {
    return this.value;
  }

  /**
   * Set the type of the constructor argument.
   */
  public set Type(type: string) {
    this.type = type;
  }

  /**
   * Return the type of the constructor argument.
   */

  public get Type(): string {
    return this.type;
  }

  /**
   * Set the name of the constructor argument.
   */
  public set Name(name: string) {
    this.name = name;
  }

  /**
   * Return the name of the constructor argument.
   */

  public get Name(): string {
    return this.name;
  }

  /**
   * Set the configuration source {@code Object} for this metadata element.
   * <p>The exact type of the object will depend on the configuration mechanism used.
   */
  public set Source(source: any) {
    this.source = source;
  }

  public get Source(): any {
    return this.source;
  }

  /**
   * Return whether this holder contains a converted value already ({@code true}),
   * or whether the value still needs to be converted ({@code false}).
   */
  public isConverted(): boolean {
    return this.converted;
  }

  /**
   * Set the converted value of the constructor argument,
   * after processed type conversion.
   */
  public set ConvertedValue(value: any) {
    this.converted = (value != null);
    this.convertedValue = value;
  }

  /**
   * Return the converted value of the constructor argument,
   * after processed type conversion.
   */
  public get ConvertedValue(): any {
    return this.convertedValue;
  }

  /**
   * Determine whether the content of this ValueHolder is equal
   * to the content of the given other ValueHolder.
   * <p>Note that ValueHolder does not implement {@code equals}
   * directly, to allow for multiple ValueHolder instances with the
   * same content to reside in the same Set.
   */
  private contentEquals(other: ValueHolder): boolean {
    return (this === other ||
        (ObjectUtils.nullSafeEquals(this.value, other.value) && ObjectUtils.nullSafeEquals(this.type, other.type)));
  }

  /**
   * Determine whether the hash code of the content of this ValueHolder.
   * <p>Note that ValueHolder does not implement {@code hashCode}
   * directly, to allow for multiple ValueHolder instances with the
   * same content to reside in the same Set.
   */
  private contentHashCode(): number {
    return ObjectUtils.nullSafeHashCode(this.value) * 29 + ObjectUtils.nullSafeHashCode(this.type);
  }

  /**
   * Create a copy of this ValueHolder: that is, an independent
   * ValueHolder instance with the same contents.
   */
  public copy(): ValueHolder {
    ValueHolder; copy = new ValueHolder(this.value, this.type, this.name);
    copy.setSource(this.source);
    return copy;
  }
}


export default ConstructorArgumentValues;
